/*
 * Copyright 2018 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.util;

import io.confluent.kafka.schemaregistry.client.SchemaRegistryClient;
import io.confluent.ksql.internal.QueryStateListener;
import io.confluent.ksql.metrics.StreamsErrorCollector;
import io.confluent.ksql.planner.PlanSourceExtractorVisitor;
import io.confluent.ksql.planner.plan.OutputNode;
import io.confluent.ksql.serde.DataSource;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import org.apache.kafka.connect.data.Schema;
import org.apache.kafka.streams.KafkaStreams;
import org.apache.kafka.streams.Topology;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class QueryMetadata {

  private static final Logger log = LoggerFactory.getLogger(QueryMetadata.class);

  private final String statementString;
  private final KafkaStreams kafkaStreams;
  private final OutputNode outputNode;
  private final String executionPlan;
  private final DataSource.DataSourceType dataSourceType;
  private final String queryApplicationId;
  private final KafkaTopicClient kafkaTopicClient;
  private final Topology topoplogy;
  private final Map<String, Object> overriddenProperties;
  private final Set<String> sourceNames;

  private Optional<QueryStateListener> queryStateListener = Optional.empty();

  protected QueryMetadata(final String statementString,
                       final KafkaStreams kafkaStreams,
                       final OutputNode outputNode,
                       final String executionPlan,
                       final DataSource.DataSourceType dataSourceType,
                       final String queryApplicationId,
                       final KafkaTopicClient kafkaTopicClient,
                       final Topology topoplogy,
                       final Map<String, Object> overriddenProperties) {
    this.statementString = statementString;
    this.kafkaStreams = kafkaStreams;
    this.outputNode = outputNode;
    this.executionPlan = executionPlan;
    this.dataSourceType = dataSourceType;
    this.queryApplicationId = queryApplicationId;
    this.kafkaTopicClient = kafkaTopicClient;
    this.topoplogy = topoplogy;
    this.overriddenProperties = overriddenProperties;
    final PlanSourceExtractorVisitor<?, ?> visitor = new PlanSourceExtractorVisitor<>();
    visitor.process(outputNode, null);
    this.sourceNames = visitor.getSourceNames();
  }

  public void registerQueryStateListener(final QueryStateListener queryStateListener) {
    this.queryStateListener = Optional.of(queryStateListener);
  }

  public Map<String, Object> getOverriddenProperties() {
    return overriddenProperties;
  }

  public String getStatementString() {
    return statementString;
  }

  public KafkaStreams getKafkaStreams() {
    return kafkaStreams;
  }

  public OutputNode getOutputNode() {
    return outputNode;
  }

  public String getExecutionPlan() {
    return executionPlan;
  }

  public DataSource.DataSourceType getDataSourceType() {
    return dataSourceType;
  }

  public String getQueryApplicationId() {
    return queryApplicationId;
  }

  public Topology getTopology() {
    return topoplogy;
  }

  public Schema getResultSchema() {
    return outputNode.getSchema();
  }

  public Set<String> getSourceNames() {
    return sourceNames;
  }


  /**
   * Stops the query without cleaning up the external resources
   * so that it can be resumed when we call {@link #start()}.
   *
   * <p>NOTE: {@link QueuedQueryMetadata} overrides this method
   * since any time a transient query is stopped the external resources
   * should be cleaned up.</p>
   *
   * @see #close()
   */
  public abstract void stop();

  /**
   * Closes the {@code QueryMetadata} and cleans up any of
   * the resources associated with it (e.g. internal topics,
   * schemas, etc...).
   *
   * @see QueryMetadata#stop()
   */
  public void close() {
    doClose(true);
  }

  protected void doClose(final boolean cleanUp) {
    kafkaStreams.close();
    if (cleanUp) {
      kafkaStreams.cleanUp();
      kafkaTopicClient.deleteInternalTopics(queryApplicationId);
    }
    queryStateListener.ifPresent(QueryStateListener::close);
    StreamsErrorCollector.notifyApplicationClose(queryApplicationId);
  }

  private Set<String> getInternalSubjectNameSet(final SchemaRegistryClient schemaRegistryClient) {
    try {
      final String suffix1 = KsqlConstants.STREAMS_CHANGELOG_TOPIC_SUFFIX
                             + KsqlConstants.SCHEMA_REGISTRY_VALUE_SUFFIX;
      final String suffix2 = KsqlConstants.STREAMS_REPARTITION_TOPIC_SUFFIX
                             + KsqlConstants.SCHEMA_REGISTRY_VALUE_SUFFIX;

      return schemaRegistryClient.getAllSubjects().stream()
          .filter(subjectName -> subjectName.startsWith(getQueryApplicationId()))
          .filter(subjectName -> subjectName.endsWith(suffix1) || subjectName.endsWith(suffix2))
          .collect(Collectors.toSet());
    } catch (final Exception e) {
      // Do nothing! Schema registry clean up is best effort!
      log.warn("Could not clean up the schema registry for query: " + queryApplicationId, e);
    }
    return new HashSet<>();
  }


  public void cleanUpInternalTopicAvroSchemas(final SchemaRegistryClient schemaRegistryClient) {
    getInternalSubjectNameSet(schemaRegistryClient).forEach(subjectName -> {
      try {
        schemaRegistryClient.deleteSubject(subjectName);
      } catch (final Exception e) {
        log.warn("Could not clean up the schema registry for query: " + queryApplicationId
                 + ", topic: " + subjectName, e);
      }
    });
  }

  @Override
  public boolean equals(final Object o) {
    if (!(o instanceof QueryMetadata)) {
      return false;
    }

    final QueryMetadata that = (QueryMetadata) o;

    return Objects.equals(this.statementString, that.statementString)
        && Objects.equals(this.kafkaStreams, that.kafkaStreams)
        && Objects.equals(this.outputNode, that.outputNode)
        && Objects.equals(this.queryApplicationId, that.queryApplicationId);
  }

  @Override
  public int hashCode() {
    return Objects.hash(statementString, kafkaStreams, outputNode, queryApplicationId);
  }

  public void start() {
    log.info("Starting query with application id: {}", queryApplicationId);
    queryStateListener.ifPresent(kafkaStreams::setStateListener);
    kafkaStreams.start();
  }

  public String getTopologyDescription() {
    return topoplogy.describe().toString();
  }
}
